import inspect
import os
import time
from enum import Enum
from functools import partial
from queue import Queue
from threading import Thread, Lock
from typing import Optional, Generator, Any, Union, \
    MutableSequence, Sequence, Tuple, List, Callable, Iterable
from urllib.parse import urljoin, urlparse

import requests
from requests.exceptions import ConnectionError, Timeout

from .exceptions import (
    SpiderError,
    RequestIgnored,
    DownLoadError,
    ResponseIgnored,
    ParseError,
    PipeError,
    FailedStatusCode,
)
from .middlewares import random_useragent
from .request import Request
from .response import Response
from .utils import (
    logger,
    DelayQueue,
    random_range,
    identity,
    assert_positive_integer,
    assert_not_negative_integer,
    assert_positive_number,
    assert_not_negative_number,
)

__all__ = [
    'Spider',
    'before_download',
    'after_download',
    'pipe',
]


class HandlerType(Enum):
    BEFORE_DOWNLOAD = 1
    AFTER_DOWNLOAD = 2
    PIPE = 3


def before_download(func):
    func._ht = HandlerType.BEFORE_DOWNLOAD
    return func


def after_download(func):
    func._ht = HandlerType.AFTER_DOWNLOAD
    return func


def pipe(func):
    func._ht = HandlerType.PIPE
    return func


class Spider:
    # The number of concurrent requests that will be performed by the downloader.
    WORKERS = os.cpu_count() * 2

    # The amount of time (in secs) that the downloader will wait before timeout.
    TIMEOUT = 30

    # The amount of time (in secs) that the downloader should wait before download.
    DOWNLOAD_DELAY = 0

    # If enabled, the downloader will wait a random time (0.5 * delay ~ 1.5 * delay by default)
    # before downloading the next page.
    RANDOM_DOWNLOAD_DELAY = True

    # Maximum number of times to retry when encountering connection issues or unexpected status codes.
    RETRY_TIMES = 3

    # HTTP response status codes to retry.
    # Other errors (DNS or connection issues) are always retried.
    # 502: Bad Gateway, 503: Service Unavailable, 504: Gateway Timeout
    # 408: Request Timeout, 429: Too Many Requests
    RETRY_CODES = (500, 502, 503, 504, 408, 429)

    # The amount of time (in secs) that the downloader will wait before retrying a failed request.
    RETRY_DELAY = 1

    MAX_REQUEST_QUEUE_SIZE = 256

    DEFAULT_HEADERS = {
        'User-Agent': 'mocy/0.1 (a kind spider)',
        'Cache-Control': 'max-age=0',
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
        'Accept-Language': 'accept-language: zh-Hans-CN,zh-CN;q=0.9,zh;q=0.8,en;q=0.7,en-GB;q=0.6,en-US;q=0.5,'
                           'zh-TW;q=0.4,ja;q=0.3,pt;q=0.2,hu;q=0.1',

    }

    before_download_handlers: List[Callable] = [random_useragent]

    after_download_handlers: List[Callable] = []

    pipes: List[Callable] = []

    entry: Union[str, Request, Iterable[Union[str, Request]], Callable] = []

    def __init__(self) -> None:
        self._check_config()
        self._check_handlers()

        self._request_queue = DelayQueue(self.MAX_REQUEST_QUEUE_SIZE)
        self._response_queue = Queue()

        self._request_count = 0
        self._response_count = 0
        self._failed_urls = []

        self._last_download_time = 0
        self._lock = Lock()

    def on_start(self) -> None:
        """Called when the spider starts up."""

    def parse(self, res: Response) -> Any:
        """Parse a response and generate some data or new requests."""
        yield res

    def collect(self, item: Any) -> Any:
        """Called when the spider outputs a result.
        Usually it will be called multiple times。"""
        logger.info(item)

    def on_finish(self) -> None:
        """Called when the spider exits."""

    def on_error(self, reason: SpiderError) -> None:
        """Called when the spider encounters an error when downloading or parsing.
        It may be called multiple times。"""
        logger.error(reason.msg, exc_info=reason.cause)

    def start(self) -> None:
        """Starts up the spider.
        It will keep running until all requests were downloaded."""
        start = time.time()
        logger.info('Spider is running...')

        for item in self:
            error = None
            if not isinstance(item, SpiderError):
                try:
                    self._start_pipes(*item)
                except Exception as err:
                    req, res = item[1].req, item[1]
                    error = PipeError(req.url, err)
                    error.req, error.res = req, res
            else:
                error = item

            if not error:
                continue

            try:
                self.on_error(error)
            except Exception as err:
                logger.error('Error in error handler!', exc_info=err)

        logger.info('Spider exited; total running time {:.2f}s.'.format(time.time() - start))
        self._log_failed_urls()

    def __iter__(self) -> Generator[Union[SpiderError, Tuple[Any, requests.Response]], None, None]:
        self.on_start()
        self._start_requests()
        self._start_downloaders()

        while not self._completed:
            res = self._response_queue.get()
            self._response_count += 1

            if isinstance(res, SpiderError):
                err = self._check_error(res)
                if err:
                    yield err
                continue

            res.select = partial(Response.select, res)

            parse = getattr(res.req, 'callback') or self.parse

            session = res.session
            close_session = True

            try:
                result = parse(res)
                if isinstance(result, (MutableSequence, Generator)):
                    for item in result:
                        if isinstance(item, Request):
                            req = item
                            req.url = urljoin(res.url, req.url)

                            if session and (req.session in (False, None)):
                                close_session = False
                                req.session = session

                            req.headers['Referer'] = res.url
                            self._add_request(req)
                        else:
                            yield item, res
            except Exception as err:
                err = ParseError(res.req.url, err)
                err.req = res.req
                err.res = res
                yield err

            if session and close_session:
                session.close()

        self.on_finish()

    @classmethod
    def _check_config(cls) -> None:
        assert_positive_integer(cls.WORKERS)
        assert_positive_integer(cls.MAX_REQUEST_QUEUE_SIZE)
        assert_not_negative_integer(cls.RETRY_TIMES)
        assert_not_negative_number(cls.RETRY_DELAY)
        assert_not_negative_number(cls.DOWNLOAD_DELAY)

        for code in cls.RETRY_CODES:
            assert code in range(400, 600)

        timeout = cls.TIMEOUT
        if isinstance(timeout, Sequence):
            assert len(timeout) == 2
            for num in timeout:
                assert_positive_number(num)
        else:
            assert_positive_number(timeout)

        random_delay = cls.RANDOM_DOWNLOAD_DELAY
        assert isinstance(random_delay, (bool, Sequence))
        if isinstance(random_delay, Sequence):
            assert len(random_delay) == 2
            for num in random_delay:
                assert_positive_number(num)

    @classmethod
    def _check_handlers(cls) -> None:
        def check(handlers, name):
            for idx, handler in enumerate(handlers):
                if inspect.isfunction(handler):
                    return
                if hasattr(handler, name):
                    hdl = getattr(handler, name)
                else:
                    logger.warn('No `{}` handler in {}'.format(name, handler.__class__))
                    hdl = identity

                handlers[idx] = hdl

        check(cls.before_download_handlers, 'before_download')
        check(cls.after_download_handlers, 'after_download')

    def _start_requests(self) -> None:
        entry = self.entry
        if inspect.ismethod(entry):
            entry = entry()

        if isinstance(entry, (str, Request)):
            entry = [entry]

        for req in entry:
            assert isinstance(req, (str, Request))
            if not isinstance(req, Request):
                req = Request(req)
            self._add_request(req)

    def _add_request(self, req: 'Request') -> None:
        self._request_count += 1

        if req.timeout is None:
            req.timeout = self.TIMEOUT

        self._add_default_header(req)

        if req.retry_num > 0 and self.RETRY_DELAY > 0:
            self._request_queue.put_later(req, self.RETRY_DELAY)
        else:
            self._request_queue.put(req)

    def _add_default_header(self, req: Request) -> None:
        headers = req.headers
        headers.setdefault('Host', urlparse(req.url).netloc)
        for name, value in self.DEFAULT_HEADERS.items():
            headers.setdefault(name, value)

    def _start_downloaders(self) -> None:
        for idx in range(self.WORKERS):
            thread = Thread(target=self._download, name='downloader-{}'.format(idx))
            thread.daemon = True
            thread.start()

    def _download(self) -> None:
        while True:
            req = self._request_queue.get()

            self._wait(self._get_download_delay())

            # before download
            try:
                req = self._pre_download(req)
            except Exception as err:
                if not isinstance(err, RequestIgnored):
                    err = RequestIgnored(req.url, err)
                err.req = req
                self._response_queue.put(err)
                continue

            # downloading
            try:
                t0 = time.time()
                res = req.send()
                t1 = time.time()
                logger.info('"{} {}" {} {:.2f}s'.format(
                    req.method, req.url, res.status_code, t1 - t0
                ))
                self._check_status_codes(res)

            except Exception as err:
                if not isinstance(err, DownLoadError):
                    err = DownLoadError(req.url, err)
                if isinstance(err.cause, (ConnectionError, Timeout, FailedStatusCode)):
                    err.need_retry = True
                err.req = req
                self._response_queue.put(err)
                continue

            # after download
            try:
                res = self._post_download(res)
            except Exception as err:
                if not isinstance(err, ResponseIgnored):
                    err = ResponseIgnored(res.req.url, err)
                err.req = req
                err.res = res
                self._response_queue.put(err)
                continue

            self._response_queue.put(res)

    @classmethod
    def _check_status_codes(cls, res: Response) -> None:
        if res.status_code in cls.RETRY_CODES:
            err = DownLoadError(res.req.url, FailedStatusCode(res.status_code))
            err.res = res
            raise err

    def _wait(self, delay) -> None:
        if delay <= 0:
            return

        with self._lock:
            rest = time.time() - self._last_download_time
            if delay > rest:
                time.sleep(delay - rest)
            self._last_download_time = time.time()

    @classmethod
    def _get_download_delay(cls) -> Union[int, float]:
        delay = cls.DOWNLOAD_DELAY
        if delay <= 0:
            return 0

        random = cls.RANDOM_DOWNLOAD_DELAY

        if not random:
            return delay

        if random is True:
            s1, s2 = 0.5, 1.5
        else:
            s1, s2 = random

        return random_range(delay, s1, s2)

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        if vars(cls).get('before_download_handlers') is None:
            cls.before_download_handlers = Spider.before_download_handlers.copy()
        if vars(cls).get('after_download_handlers') is None:
            cls.after_download_handlers = Spider.after_download_handlers.copy()
        if vars(cls).get('pipes') is None:
            cls.pipes = Spider.pipes.copy()

        # class attribute definition order is preserved from 3.6
        for value in cls.__dict__.values():
            if inspect.isfunction(value) and hasattr(value, '_ht'):
                ht = getattr(value, '_ht')
                if ht == HandlerType.BEFORE_DOWNLOAD:
                    cls.before_download_handlers.append(value)
                elif ht == HandlerType.AFTER_DOWNLOAD:
                    cls.after_download_handlers.append(value)
                else:
                    cls.pipes.append(value)

    def _pre_download(self, req: 'Request') -> Optional['Request']:
        rv = req
        for handler in self.before_download_handlers:
            rv = handler(self, rv)
            if not isinstance(rv, Request):
                raise RequestIgnored(req.url)
        return rv

    def _post_download(self, res: 'Response') -> Optional[Union['Response', 'Request']]:
        rv = res
        for handler in self.after_download_handlers:
            rv = handler(self, rv)
            if not isinstance(rv, requests.Response):
                err = ResponseIgnored(res.req.url)
                if isinstance(rv, Request):
                    err.new_req = rv
                raise err
        return rv

    def _start_pipes(self, item: Any, res: 'Response') -> Any:
        rv = item
        for func in (self.pipes or (self.__class__.collect,)):
            arg_count = func.__code__.co_argcount
            if arg_count == 3:
                rv = func(self, rv, res)
            else:
                rv = func(self, rv)

            if rv is None:
                return None
        return rv

    def _check_error(self, error: SpiderError) -> Optional[SpiderError]:
        if isinstance(error, RequestIgnored):
            if error.cause:
                return error
            else:
                logger.debug(f'Request for {error.req.url} was ignored')
        elif isinstance(error, ResponseIgnored):
            if error.new_req:
                self._add_request(error.new_req)
            if error.cause:
                return error
            else:
                logger.debug(f'Response for {error.req.url} was ignored')
        elif isinstance(error, DownLoadError):
            req = error.req
            if error.need_retry and req.retry_num < self.RETRY_TIMES:
                req.retry_num += 1
                logger.debug('Retrying ({}) for {}...'.format(req.retry_num, req.url))
                self._add_request(req)
            else:
                self._failed_urls.append(req.url)
                return error

    def _log_failed_urls(self) -> None:
        urls = self._failed_urls
        if not urls:
            return

        num = len(urls)
        s = 's' if num > 1 else ''
        msg = 'Fail to download from the following {} url{}:\n{}'.format(num, s, '\n'.join(urls))
        logger.error(msg)

    @property
    def _completed(self) -> bool:
        return self._request_count == self._response_count



import logging
import re
import sys
import time
from queue import Queue, PriorityQueue
from random import random, randint
from threading import Thread
from typing import Union
from urllib.parse import urlparse

__all__ = [
    'DelayQueue',
    'logger',
    'random_range',
    'random_ip',
    'same_origin',
    'identity',
    'assert_positive_number',
    'assert_not_negative_number',
    'assert_positive_integer',
    'assert_not_negative_integer',
]


class DelayQueue(Queue):
    def __init__(self, maxsize=0):
        super().__init__(maxsize)
        self.pq = PriorityQueue()
        poller = Thread(target=self._poll, name='poller')
        poller.daemon = True
        poller.start()

    def put_later(self, item, delay=1):
        self.pq.put((time.time() + delay, item))

    def _poll(self):
        while True:
            item = self.pq.get()
            if item[0] <= time.time():
                self.put(item[1])
            else:
                self.pq.put(item)
                time.sleep(0.05)


def random_range(
        value: Union[int, float],
        scale1: Union[int, float],
        scale2: Union[int, float]
) -> float:
    if scale1 > scale2:
        lo, hi = scale2, scale1
    else:
        lo, hi = scale1, scale2
    factor = lo + (hi - lo) * random()
    return factor * value


def identity(x):
    return x


def assert_not_negative_number(num):
    assert num >= 0 and isinstance(num, (int, float))


def assert_positive_number(num):
    assert num > 0 and isinstance(num, (int, float))


def assert_not_negative_integer(num):
    assert num >= 0 and isinstance(num, int)


def assert_positive_integer(num):
    assert num > 0 and isinstance(num, int)


def add_http_if_no_scheme(url: str) -> str:
    """Add http as the default scheme if it is missing from the url."""
    match = re.match(r'^\w+://', url, flags=re.I)
    if not match:
        parts = urlparse(url)
        scheme = "http:" if parts.netloc else "http://"
        url = scheme + url
    return url


def same_origin(url1: str, url2: str) -> bool:
    """Return True if the two urls are the same origin
    >>> same_origin('http://a.com', 'https://a.com')
    False
    >>> same_origin('https://a.com', 'https://a.com:8080')
    False
    >>> same_origin('https://a.com/foo', 'https://a.com/bar')
    True
    """
    return all(map(
        lambda x: x[0] == x[1],
        list(zip(urlparse(url1), urlparse(url2)))[0:2]
    ))


def random_ip() -> str:
    """A simple ipv4 generator that filters some special ips."""
    specials = [0, 10, 100, 127, 172, 192, 198, 203, 224, 240, 255]

    def gen(): return randint(0, 255)

    while True:
        prefix = gen()
        if prefix not in specials:
            break
    return '.'.join(map(str, (prefix, gen(), gen(), gen())))


class Logger:
    logger_format = '[%(asctime)-15s] %(levelname)-7s: %(message)s'
    date_format = '%Y-%m-%d %H:%M:%S'

    def __init__(self, name: str, level: int = logging.DEBUG) -> None:
        self._logger = logging.getLogger(name)
        self._logger.setLevel(level)
        self._add_stream_handlers()

    def replace(self, new_logger: logging.Logger) -> None:
        self._logger = new_logger

    def set_level(self, level: int) -> None:
        self._logger.setLevel(level)

    def add_handler(self, handler: logging.Handler) -> None:
        self._logger.addHandler(handler)

    def debug(self, msg: str, *args, **kwargs) -> None:
        self._logger.debug(msg, *args, **kwargs)

    def info(self, msg: str, *args, **kwargs) -> None:
        self._logger.info(msg, *args, **kwargs)

    def warn(self, msg: str, *args, **kwargs) -> None:
        self._logger.warning(msg, *args, **kwargs)

    def error(self, msg: str, *args, **kwargs) -> None:
        self._logger.error(msg, *args, **kwargs)

    def _add_stream_handlers(self):
        stdout_handler = self._stream_handler(
            sys.stdout,
            logging.DEBUG,
            lambda record: record.levelno < logging.ERROR
        )
        stderr_handler = self._stream_handler(
            sys.stderr,
            logging.ERROR
        )
        self._logger.addHandler(stdout_handler)
        self._logger.addHandler(stderr_handler)

    def _stream_handler(self, stream, level, msg_filter=None):
        handler = logging.StreamHandler(stream)
        handler.setLevel(level)
        formatter = logging.Formatter(self.logger_format, datefmt=self.date_format)
        handler.setFormatter(formatter)
        if msg_filter:
            handler.addFilter(msg_filter)
        return handler


logger = Logger('mocy')



from numbers import Number
from typing import Optional, Union, Callable, Tuple

import requests


class Request:
    def __init__(self,
                 url: str,
                 method: str = 'GET',

                 callback: Optional[Callable] = None,
                 session: Union[bool, dict, requests.Session] = False,
                 state: Optional[dict] = None,

                 headers: Optional[dict] = None,
                 cookies: Optional[dict] = None,
                 params: Optional[dict] = None,
                 data: Optional[dict] = None,
                 json: Optional[dict] = None,
                 files: Optional[dict] = None,
                 proxies: Optional[dict] = None,
                 verify: bool = True,
                 timeout: Optional[Union[Tuple[Number, Number], Number]] = None,
                 **kwargs) -> None:
        self.url = url
        self.method = method

        self.callback = callback
        self.session = session
        self.state = state

        self.headers = headers or {}
        self.cookies = cookies
        self.params = params
        self.data = data
        self.json = json
        self.files = files
        self.proxies = proxies
        self.verify = verify
        self.timeout = timeout

        self.kwargs = kwargs

        self.retry_num = 0

    def send(self) -> 'Response':
        """Send a request and return a response."""
        it = requests
        sess = self._get_session()
        if sess:
            it = sess

        res = it.request(self.method, self.url, **self._prepare_args())
        res.req = self
        res.state = self.state
        res.session = sess
        return res

    @property
    def initial(self) -> bool:
        """Whether it is an independent request or the first request in a session."""
        return not isinstance(self.session, requests.Session)

    def _get_session(self) -> Optional[requests.Session]:
        session = self.session
        if session is True:
            return requests.Session()
        elif isinstance(session, requests.Session):
            return session
        elif isinstance(session, dict):
            sess = requests.Session()
            for key, value in session.items():
                setattr(sess, key, value)
            return sess
        else:
            return None

    def _prepare_args(self) -> dict:
        args = {}

        for name in ('headers', 'cookies', 'params', 'data',
                     'json', 'files', 'proxies', 'verify', 'timeout'):
            value = getattr(self, name)
            if value:
                args[name] = value

        args.update(self.kwargs)
        return args

    def __repr__(self) -> str:
        return '<Request [{}]>'.format(self.method)


from .response import Response



from typing import Optional


class SpiderError(Exception):
    def __init__(self, msg: str, cause: Optional[Exception] = None) -> None:
        self.msg = msg
        self.cause = cause
        self.req = None
        self.res = None


class RequestIgnored(SpiderError):
    def __init__(self, url: str, cause: Optional[Exception] = None) -> None:
        super().__init__('Request was ignored for {}.'.format(url), cause)


class ResponseIgnored(SpiderError):
    def __init__(self, url: str, cause: Optional[Exception] = None) -> None:
        self.new_req = None
        super().__init__('Response was ignored for {}'.format(url), cause)


class DownLoadError(SpiderError):
    def __init__(self, url: str, cause: Optional[Exception] = None) -> None:
        self.need_retry = False
        super().__init__('Cannot download from {}'.format(url), cause)


class ParseError(SpiderError):
    def __init__(self, url: str, cause: Optional[Exception] = None) -> None:
        super().__init__('Error when parsing response from {}'.format(url), cause)


class PipeError(SpiderError):
    def __init__(self, url: str, cause: Optional[Exception] = None) -> None:
        super().__init__('Error when collecting results from {}'.format(url), cause)


class FailedStatusCode(Exception):
    def __init__(self, status_code: int):
        self.status_code = status_code
        super().__init__(status_code)



from typing import Optional, List

import bs4
import requests
from bs4 import BeautifulSoup

from .request import Request


try:
    import lxml
    parser = 'lxml'
except ModuleNotFoundError:
    parser = 'html.parser'


class Response(requests.Response):
    def __init__(self):
        super().__init__()
        self.req: Optional[Request] = None
        self.state: Optional[dict] = None
        self.session: Optional[requests.Session] = None

    def select(self, selector: str, **kw) -> List[bs4.element.Tag]:
        soup = BeautifulSoup(self.text, parser)
        return soup.select(selector, **kw)



import io
from setuptools import setup
import mocy

with io.open('README.md', 'rt') as f:
    readme = f.read()

setup(
    name='mocy',
    version=mocy.__version__,
    url='https://github.com/cymoo/mocy',
    description='A lightweight web crawling framework.',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='cymoo',
    author_email='wakenee@hotmail.com',
    license='MIT',
    platforms='any',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Utilities',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='scrapy crawler spider',
    packages=['mocy', 'mocy.middlewares'],
    package_data={'mocy.middlewares': ['*.txt']},
    python_requires='>=3.6',
    install_requires=[
        'requests>=2.25.1',
        'beautifulsoup4>=4.9.3'
    ],
    extras_require={'dev': ['pytest']},
)



import os
import random

dirname = os.path.dirname(os.path.abspath(__file__))

# The useragent list was taken from: https://techblog.willshouse.com/2012/01/03/most-common-user-agents/
with open(os.path.join(dirname, 'useragents.txt'), 'rt') as fp:
    useragents = []
    for line in fp:
        line = line.strip()
        if line: useragents.append(line)


def random_useragent(spider, request):
    if request.initial:
        request.headers['User-Agent'] = random.choice(useragents)
    return request
